/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2004-2006 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2 of the
 * License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <errno.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>

/* FIXME stupid thing */
#define inline

#include <bluetooth/bluetooth.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>
#include <bluetooth/rfcomm.h>

#include <gw-obex.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include <dbus/dbus-glib-lowlevel.h>

#include "om-dbus.h"

#define d(x) x

/* Gwcond communication. Talks to gwcond to get the device.
 *
 * This API can be accessed from any thread, since it creates it's own context
 * and uses that for the dbus connection.
 */

typedef struct {
	DBusConnection *dbus_conn;
	GMainContext   *context;
	GMainLoop      *loop;
} Connection;

typedef gboolean (* ForeachAdapterFunc) (Connection *conn,
					 const char *adapter,
					 gpointer user_data);

static gboolean is_obex_device (Connection *conn, 
				const char *dev,
				const char *name);

static Connection *
get_system_bus_connection (void)
{
	DBusConnection *dbus_conn;
	Connection     *conn;
	DBusError       error;

	/* NOTE: We are using dbus_bus_get_private here, for the reason that
	 * need to get our own private dbus connection to avoid threading
	 * problems with other libraries or applications that use this module
	 * and dbus (the vfs daemon in particular).
	 */
	dbus_error_init (&error);
        dbus_conn = dbus_bus_get_private (DBUS_BUS_SYSTEM, &error);	
	
	if (!dbus_conn) {
		g_printerr ("Failed to connect to the D-BUS daemon: %s", error.message);
		
		dbus_error_free (&error);
		return NULL;
	}

	conn = g_new0 (Connection, 1);
	
	conn->context = g_main_context_new ();
	conn->loop = g_main_loop_new (conn->context, FALSE);

	conn->dbus_conn = dbus_conn;

	dbus_connection_setup_with_g_main (dbus_conn, conn->context);
	
	return conn;
}

static void
connection_free (Connection *conn)
{
	dbus_connection_close (conn->dbus_conn);
	dbus_connection_unref (conn->dbus_conn);
	
	g_main_loop_unref (conn->loop);
	g_main_context_unref (conn->context);
	
	g_free (conn);
}

static gboolean
foreach_adapter (Connection *conn, ForeachAdapterFunc func, gpointer user_data)
{
	DBusMessage     *msg;
	DBusMessage     *ret;
        DBusMessageIter  iter;
	DBusError        error;

	msg = dbus_message_new_method_call ("org.bluez", "/org/bluez",
                                            "org.bluez.Manager",
                                            "ListAdapters");

	if (!msg) {
		return FALSE;
	}

	dbus_error_init (&error);

	ret = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
							 msg, -1, &error);
	dbus_message_unref (msg);

	if (dbus_error_is_set (&error)) {
		dbus_error_free (&error);
		return FALSE;
	}

	if (dbus_message_iter_init (ret, &iter)) {
                DBusMessageIter sub;

		dbus_message_iter_recurse (&iter, &sub);
		
		/* Go through each entry (device) and get each paired device
		 * from the entry.
		 */
		do {
	                char *devname;

			dbus_message_iter_get_basic (&sub, &devname);

			if (!func (conn, devname, user_data)) {
				dbus_message_unref (ret);
				return FALSE;
			}
		} while (dbus_message_iter_next (&sub));
	}

	dbus_message_unref (ret);

	return TRUE;
}

static gboolean
check_bda (const gchar *bda)
{
	gint len, i;

	if (!bda) {
		return FALSE;
	}

	len = strlen (bda);
	if (len != 17) {
		return FALSE;
	}
	
	for (i = 0; i < 17; i += 3) {
		if (!isxdigit (bda[i])) {
			return FALSE;
		}
		if (!isxdigit (bda[i+1])) {
			return FALSE;
		}
		if (i < 15 && bda[i+2] != ':') {
			return FALSE;
		}
	}
	
	return TRUE;
}

static int
connect_rfcomm (const char *source_bdaddr, const char *dest_bdaddr, uint8_t channel)
{
	struct sockaddr_rc addr;
	int fd;

	fd = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
	if (fd < 0)
		return -1;

	memset(&addr, 0, sizeof(addr));
	addr.rc_family  = AF_BLUETOOTH;

	str2ba(source_bdaddr, &addr.rc_bdaddr);
	addr.rc_channel = 0;

	if (bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		d(g_printerr ("connect_rfcomm can't bind: %s", g_strerror (errno)));
		return -1;
	}

	/* We don't set the file descriptor non-blocking otherwise we'd have
	 * to have a callback to check on the progress of the operation, I don't
	 * fancy doing that... */

	addr.rc_family  = AF_BLUETOOTH;
	str2ba(dest_bdaddr, &addr.rc_bdaddr);
	addr.rc_channel = channel;

	if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
		d(g_printerr ("connect_rfcomm can't connect: %s", g_strerror (errno)));
		close (fd);
		return -1;
	}

	return fd;
}

static char *
get_adapter_bdaddr (Connection *conn,
		    const char *adapter)
{
	DBusMessage *message = NULL;
	DBusMessage *reply = NULL;
	DBusError    dbus_error;
	char        *bdaddr, *str;

	message = dbus_message_new_method_call ("org.bluez",
						adapter,
						"org.bluez.Adapter",
						"GetAddress");

	if (!message)
		return NULL;

	dbus_error_init (&dbus_error);
	reply = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
							   message, -1,
							   &dbus_error);

	dbus_message_unref (message);

	if (dbus_error_is_set (&dbus_error)) {
		if (reply)
			dbus_message_unref (reply);
		d(g_printerr ("Couldn't get the bdaddr for source '%s': %s\n", adapter, dbus_error.name));
		dbus_error_free (&dbus_error);
		return NULL;
	}

	if (!reply) {
		d(g_printerr ("Couldn't get the bdaddr for source '%s'\n", adapter));
		return NULL;
	}

	if (!dbus_message_get_args (reply, NULL,
				    DBUS_TYPE_STRING, &str,
				    DBUS_TYPE_INVALID)) {
		dbus_message_unref (reply);
		return NULL;
	}

	bdaddr = g_strdup (str);
	dbus_message_unref (reply);

	return bdaddr;
}

/* Stolen from bluez-utils/cups/sdp.c */
static int
sdp_search_obexftp (sdp_session_t *sdp,
		    uint8_t       *channel)
{
	sdp_list_t *srch, *attrs, *rsp;
	uuid_t svclass;
	uint16_t attr;
	int err;

	if (!sdp)
		return -1;

	sdp_uuid16_create (&svclass, OBEX_FILETRANS_SVCLASS_ID);
	srch = sdp_list_append (NULL, &svclass);

	attr = SDP_ATTR_PROTO_DESC_LIST;
	attrs = sdp_list_append (NULL, &attr);

	err = sdp_service_search_attr_req (sdp, srch, SDP_ATTR_REQ_INDIVIDUAL, attrs, &rsp);
	if (err)
		return -1;

	for (; rsp; rsp = rsp->next) {
		sdp_record_t *rec = (sdp_record_t *) rsp->data;
		sdp_list_t *protos;

		if (!sdp_get_access_protos (rec, &protos)) {
			uint8_t ch = sdp_get_proto_port (protos, RFCOMM_UUID);
			if (ch > 0) {
				*channel = ch;
				return 0;
			}
		}
	}

	return -1;
}

struct get_dev_t {
	const gchar    *bda;
	GnomeVFSResult *result;
	gboolean        invalid_profile;
	gboolean        already_connected;
	int             fd;
};

/* Note: This needs to be refactored for the next version, we need our own
 * return value here, which can include invalid profile and already connected as
 * results. We can also move the bda checking to the caller, it only needs
 * checked once.
 */
static int
get_dev (Connection       *conn,
	 const gchar      *adapter,
	 struct get_dev_t *data,
	 const char       *profile)
{
	sdp_session_t *sdp;
	uint8_t        channel;
	int            fd = -1;
	bdaddr_t       source_bdaddr, dest_bdaddr;
	const char    *source_bdaddr_str;

	source_bdaddr_str = get_adapter_bdaddr (conn, adapter);
	if (source_bdaddr_str == NULL) {
		d(g_printerr("get_dev: Can't get bdaddr for adapter '%s'", adapter));
		*data->result = GNOME_VFS_ERROR_INTERNAL;
		return -1;
	}

	str2ba (source_bdaddr_str, &source_bdaddr);
	str2ba (data->bda, &dest_bdaddr);
	sdp = sdp_connect (&source_bdaddr, &dest_bdaddr, SDP_RETRY_IF_BUSY);
	if (!sdp) {
		d(g_printerr("get_dev: Can't open connection to SDP daemon"));
		g_free (source_bdaddr_str);
		*data->result = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
		return -1;
	}
	if (sdp_search_obexftp (sdp, &channel) < 0) {
		d(g_printerr("get_dev: can't get channel for '%s' on '%s'",
			     profile, data->bda));
		g_free (source_bdaddr_str);
		*data->result = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
		sdp_close(sdp);
		return -1;
	}
	sdp_close(sdp);

	d(g_printerr ("obex: Send connect to '%s' channel %d.\n", data->bda, channel));

	fd = connect_rfcomm (source_bdaddr_str, data->bda, channel);

	d(g_printerr ("obex: Sent connect to '%s' channel %d (res = %d).\n", data->bda, channel, fd));

	g_free (source_bdaddr_str);

	if (fd < 0) {
		*data->result = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
	} else {
		*data->result = GNOME_VFS_OK;
	}

	return fd;
}

#define NOKIA_FTP_UUID "00005005-0000-1000-8000-0002ee000001"

static gboolean
adapter_get_dev(Connection *conn,
		const gchar *adapter,
		gpointer user_data)
{
	struct get_dev_t *data = (struct get_dev_t *)user_data;
	const gchar *profile;

	/* From Johan Hedberg:
	 *
	 * some Nokia Symbian phones have two OBEX FTP services: one
	 * identified with the normal UUID and another with a Nokia
	 * specific 128 bit UUID. The service found behind the normal
	 * identifier is very limited in features on these phones
	 * while the other one supports full OBEX FTP (don't ask me
	 * why).
	 */
	profile = NOKIA_FTP_UUID;
	data->fd = get_dev (conn, adapter, data, profile);
	if (data->fd < 0 && data->invalid_profile) {
		profile = "FTP";
		data->fd = get_dev (conn, adapter, data, profile);
	}

	/* only continue iteration if dev is still NULL */
	return data->fd < 0;
}

int
om_dbus_get_dev (const gchar *bda, GnomeVFSResult *result)
{
	Connection      *conn;
	struct get_dev_t data;

	if (!check_bda (bda)) {
		*result = GNOME_VFS_ERROR_INVALID_URI;
		return -1;
	}

	conn = get_system_bus_connection ();
	if (!conn) {
		*result = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
		return -1;
	}

	data.bda = bda;
	data.result = result;
	data.invalid_profile = FALSE;
	data.already_connected = FALSE;
	data.fd = -1;

	foreach_adapter(conn, adapter_get_dev, (gpointer)&data);

	connection_free (conn);
	
	return data.fd;
}

static gboolean
send_disconnect (Connection *conn,
		 const gchar *adapter,
		 gpointer user_data)
{
	int fd = GPOINTER_TO_INT (user_data);

	if (fd >= 0) {
		close (fd);
		return TRUE;
	}
	return FALSE;
}

void
om_dbus_disconnect_dev (int fd)
{
	Connection *conn;

	conn = get_system_bus_connection ();
	if (!conn) {
		return;
	}

	foreach_adapter (conn, send_disconnect, GINT_TO_POINTER (fd));
	connection_free (conn);
}

struct get_display_name_t {
	const gchar    *bda;
	gchar          *name;
};

static gboolean
get_display_name (Connection *conn,
		  const char *adapter,
		  gpointer user_data)
{
	struct get_display_name_t *data;
	DBusMessage *message = NULL;
	DBusMessage *reply = NULL;
	DBusError    dbus_error;
	gboolean     ret = TRUE;
	const gchar *str = NULL;

	data = (struct get_display_name_t *)user_data;

	message = dbus_message_new_method_call ("org.bluez",
						adapter,
						"org.bluez.Adapter",
						"GetRemoteAlias");
	if (!message) {
		ret = FALSE;
		goto end;
	}
	
	if (!dbus_message_append_args (message,
				       DBUS_TYPE_STRING, &data->bda,
				       DBUS_TYPE_INVALID)) {
		ret = FALSE;
		goto end;
	}

	dbus_error_init (&dbus_error);
	reply = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
							   message, -1,
							   &dbus_error);
	if (!dbus_error_is_set (&dbus_error)) {
		/* do we have a device alias? */
		if (dbus_message_get_args (reply, NULL,
					   DBUS_TYPE_STRING, &str,
					   DBUS_TYPE_INVALID)) {
			data->name = g_strdup(str);
			ret = FALSE;
			goto end;
		}
	}

	/* clean up for next call */
	dbus_error_free (&dbus_error);
	if (reply) {
		dbus_message_unref (reply);
	}
	dbus_message_unref (message);
	message = reply = NULL;

	message = dbus_message_new_method_call ("org.bluez",
						adapter,
						"org.bluez.Adapter",
						"GetRemoteName");
	if (!message) {
		ret = FALSE;
		goto end;
	}
	
	if (!dbus_message_append_args (message,
				       DBUS_TYPE_STRING, &data->bda,
				       DBUS_TYPE_INVALID)) {
		ret = FALSE;
		goto end;
	}

	reply = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
							   message, -1,
							   &dbus_error);
	if (!dbus_error_is_set (&dbus_error)) {
		/* do we have a device alias? */
		if (dbus_message_get_args (reply, NULL,
					   DBUS_TYPE_STRING, &str,
					   DBUS_TYPE_INVALID)) {
			data->name = g_strdup(str);
			ret = FALSE;
			goto end;
		}
	}

 end:
	dbus_error_free (&dbus_error);
	if (reply) {
		dbus_message_unref (reply);
	}
	if (message) {
		dbus_message_unref (message);
	}
	return ret;
}

gchar *
om_dbus_get_display_name (const gchar    *bda,
			  GnomeVFSResult *result)
{
	Connection *conn;
	struct get_display_name_t data;

	if (!check_bda (bda)) {
		*result = GNOME_VFS_ERROR_NOT_FOUND;
		return NULL;
	}

	conn = get_system_bus_connection ();
	if (!conn) {
		*result = GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE;
		return NULL;
	}
	data.bda = bda;
	data.name = NULL;
	foreach_adapter (conn, get_display_name, (gpointer)&data);

	/* if we don't have name, use the numeric address */
	if (!data.name) {
		data.name = g_strdup (bda);
	}
	connection_free (conn);

	return data.name;
}

static void
om_append_paired_devices (Connection   *conn, 
			  DBusMessage  *msg,
                          const char   *devname, 
			  GList       **list)
{
        DBusMessageIter diter;
	DBusMessageIter dsub;

        if (!dbus_message_iter_init (msg, &diter)) {
		return;
	}

	dbus_message_iter_recurse (&diter, &dsub);
	
	do { 
		/* Add the entry to the list. */
		GnomeVFSFileInfo *info;
		char             *remote_devname;
		
		if (dbus_message_iter_get_arg_type (&dsub) != DBUS_TYPE_STRING) {
			continue;
		}
		
		dbus_message_iter_get_basic (&dsub, &remote_devname);
		
		if (!is_obex_device (conn, devname, (const char*) remote_devname)) {
			continue;
		}
		
		
		info = gnome_vfs_file_info_new ();
		
		if (!info) {
			return;
		}
		
		info->valid_fields = 
			GNOME_VFS_FILE_INFO_FIELDS_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS |
			GNOME_VFS_FILE_INFO_FIELDS_ACCESS |
			GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
			GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
		
		info->name = g_strdup (remote_devname);
		info->type = GNOME_VFS_FILE_TYPE_REGULAR;
		info->permissions = 
			GNOME_VFS_PERM_USER_READ |
			GNOME_VFS_PERM_GROUP_READ |
			GNOME_VFS_PERM_OTHER_READ |
			GNOME_VFS_PERM_ACCESS_READABLE;
		
		info->uid = 0;
		info->gid = 0;
		info->mime_type = g_strdup ("application/x-desktop");

		*list = g_list_append (*list, info);
	} while (dbus_message_iter_next (&dsub));
}

/* Leave this in for easy testing. */
#if 0
static GList *
append_fake_device (GList *list, const gchar *bda)
{
	GnomeVFSFileInfo *info;
	
	info = gnome_vfs_file_info_new ();
	
	info->valid_fields = 
		GNOME_VFS_FILE_INFO_FIELDS_TYPE |
		GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS |
		GNOME_VFS_FILE_INFO_FIELDS_MIME_TYPE |
		GNOME_VFS_FILE_INFO_FIELDS_SYMLINK_NAME;
	
	info->flags |= GNOME_VFS_FILE_FLAGS_SYMLINK;
	
	info->name = g_strdup_printf ("[%s]", bda);
	info->type = GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK;
	info->permissions = 
		GNOME_VFS_PERM_USER_READ |
		GNOME_VFS_PERM_GROUP_READ |
		GNOME_VFS_PERM_OTHER_READ;
	info->symlink_name = g_strdup_printf ("obex://[%s]", bda);
		
	info->uid = 0;
	info->gid = 0;
	info->mime_type = g_strdup ("x-directory/normal");

	/*g_print ("added fake: %s %s\n", info->name, info->symlink_name);*/

	return g_list_append (list, info);
}
#endif

static gboolean
list_bondings (Connection *conn,
	       const char *adapter,
	       gpointer user_data)
{
	GList **devlist = user_data;
	DBusMessage *msg;
	DBusMessage *ret;
	DBusError error;

	msg = dbus_message_new_method_call ("org.bluez",
					    adapter, 
					    "org.bluez.Adapter", 
					    "ListRemoteDevices");

	if (!msg) {
		return FALSE;
	}

	dbus_error_init (&error);

	ret = dbus_connection_send_with_reply_and_block (conn->dbus_conn, 
							 msg, 
							 -1, 
							 &error);
	dbus_message_unref (msg);

	if (dbus_error_is_set (&error)) {
		dbus_error_free (&error);
		return FALSE;
	}


	om_append_paired_devices (conn,
				  ret,
				  adapter,
				  devlist);

	dbus_message_unref (ret);

	return TRUE;
}

GList *
om_dbus_get_dev_list (void)
{
	Connection      *conn;
	GList           *devlist = NULL;

#if 0
	if (0) {
		devlist = append_fake_device (devlist, "foo");
		devlist = append_fake_device (devlist, "bar");
		
		return devlist;
	}
#endif
	
	conn = get_system_bus_connection ();
	if (!conn) {
		return NULL;
	}

	foreach_adapter (conn, list_bondings, (gpointer) &devlist);
	connection_free (conn);

	return devlist;
}

static gboolean
is_obex_device (Connection *conn,
		const char *dev, 
		const char *name)
{
	DBusMessage      *msg, *ret;
	DBusMessageIter  iter, sub;
	DBusError        error;
	gboolean         is_obex = FALSE;
	char            *class_name;

	msg = dbus_message_new_method_call ("org.bluez",
					    dev,
					    "org.bluez.Adapter",
					    "GetRemoteServiceClasses");

	dbus_message_iter_init_append (msg, &iter);

	dbus_message_iter_append_basic (&iter, DBUS_TYPE_STRING, &name);

	dbus_error_init (&error);
	ret = dbus_connection_send_with_reply_and_block (conn->dbus_conn,
							 msg, 
							 -1, 
							 &error);
	dbus_message_unref (msg);

	if (dbus_error_is_set (&error)) {
		dbus_error_free (&error);
		return FALSE;
	}

	if (dbus_message_iter_init (ret, &iter)) {
		dbus_message_iter_recurse (&iter, &sub);

		do {
			if (dbus_message_iter_get_arg_type (&sub) != DBUS_TYPE_STRING) {
				continue;
			}

			dbus_message_iter_get_basic (&sub, &class_name);

			if (!strcmp (class_name, "object transfer")) {
				is_obex = TRUE;
				break;
			}


		} while (dbus_message_iter_next (&sub));
	}

	dbus_message_unref (ret);

	return is_obex;
}

